//NOTE THAT THIS ALGORITHM IS A REIMPLEMENTATION OF T. QUICKS WORK, DESCRIBED
//IN A 2003 ECAL PAPER:
//Quick, T., Nehaniv, C. L., Dautenhahn, K. and Roberts, G. Evolving Embodied Genetic Regulatory Network-Driven Control Systems. In ECAL, pages 266-277, 2003.

/*
GRN algorithm in an artificial evolution framework to model biological clocks
Copyright (C) 2005  Johannes Knabe
j.f.knabe@herts.ac.uk

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
public class QuickGene implements gene{
/*
Gets the data of one Gene, i.e. it's Cis-sites, activation level,
output protein and output function
*/

public static int stdSlope=5;//slope value is different for T Quick (problem specific)
public static int stdRange=150;
public static double actiInputCeiling=1000.0;
public int[] data;
public int[][] geneCisReceptorSites;//first array: number of sites
                                    //2nd array:   value one activ/inhib,
                                    //             value two and plus ProteinSites (in Quick implementation always exactly one, but later several ones for AND combos)
public double[][] proteinSiteActivations;//first array: number of sites
                                      //2nd array:   for every protein binder the activation (in Quick implementation always exactly one binder per site but later several proteins bind per site [AND combination])
public int numberOfCisSites;
public int geneOutputProtein;
public int geneOutputActivity;

public String toString(){
String ret="";
for(int i=0;i<data.length;i++)
  ret+=data[i]+",";
return ret;
}

public QuickGRN holder;

public QuickGene(int[] data, QuickGRN holder){
this.holder=holder;
this.data=new int[data.length];
for(int i=0;i<data.length;i++)
  this.data[i]=data[i];
numberOfCisSites=1; //for Quick implementation always =1

int[] cis=new int[3];
int[] out=new int[3];

//in the Quick implementation very easy as always one 4bit cis site
System.arraycopy(data,0,cis,0,3);
geneCisReceptorSites=new int[numberOfCisSites][2];//one site, two values
geneCisReceptorSites[0][0]=data[3];//value one: inhibitory or activatory
geneCisReceptorSites[0][1]=QuickGRN.fromBinary(cis);//value two: number of matching Protein 

proteinSiteActivations=new double[geneCisReceptorSites.length][1];//one site, one value (Quick has only one binding protein per site)

System.arraycopy(data,data.length-4,out,0,3);
geneOutputProtein =QuickGRN.fromBinary(out);
geneOutputActivity=data[data.length-1];
}

/*
returns the decoded data of one cis site, specified by the param
numbero; which can be between [0,numberOfCisSites[ (like an Array)
the return value will be and int array, with:
1st value: 0 inhibitory or 1 activatory site
2nd value: Protein number which binds to this site
more values: n/a for Quick implementation but in the future this will
             be more of 2nd value, i.e. Protein number which are AND connected
*/
public int[] getCisSite(int numbero){
return geneCisReceptorSites[numbero];
}

/*
This is called one every gene to tell it, what fraction of the global 
concentration of protein "which" binds to every matching cis-site.
(NB: The the number of matching cis-sites on this gene might be 0 to n;
 but every matching site gets "strength" to bind)
[ONLY INPUT TO THE GENE EVERY CYCLE]
*/
public void bindingProtein(int which, double strength){
for(int i=0;i<geneCisReceptorSites.length;i++){
  for(int j=1;j<geneCisReceptorSites[i].length;j++){ //start from 1 as array value one is only the activ/inhib info
    if(geneCisReceptorSites[i][j]==which) //if binding matches
      proteinSiteActivations[i][j-1]+=strength;
    }
  }
}

/*
re-implementation of the activation function family derived from tanh used by Tom Quick
     # @param slope Angle of slope gets shallower as this int increases. Activation point gets earlier too
     # @param range Range of output values. E.g. if range == 4, output may be 0<=out<=4, or -2=<out<=2, or ...
     # @param floor Min output. Floor by default is minus (range/2). 
     
     def MathLib.tanhScaled(x, range, floor=nil, rshift=0, slope=2)
*/
public static double tanhScaled(double x, double range, double floor, double rshift, double slope){
//# ------------------- Move 'S' curve vertically  ------------------- #
//# floorScale is what we add at the end to shift the whole 's' curve up.
double floorScale;
if(floor<-999990.0) //# undef, so we won't tweak output up or down at all
  floorScale = .0;
else{ //# check what floor would be, and adjust to match what we want it to be
  double naturalFloor = -(range/2);
  floorScale = floor - naturalFloor;
}
//# ------------------- Move 'S' curve horizontally------------------- #
//# By reducing value of x, we effectively shift the whole 'S' to the right.
//# Need to do this, if, for example, don't want to get any output before
//# input goes over 0.0.
x = x - rshift;

return ( (((Math.exp(x/slope) - Math.exp(-x/slope)) / 2) / 
          ((Math.exp(x/slope) + Math.exp(-x/slope)) / 2)
         ) 
         * (range/2)
       ) + floorScale;
}

/*
Currently there is only one output Protein per gene, so for all except this one immediately zero will be returned. For the output protein
activations on all cis-sites are summed (+ if activatory, - if inhibitory) and put through the specific tanh derived output function.
*/
public double getOutputForProtein(int which){
if(which!=geneOutputProtein) //We generate only one output protein!
  return 0.0;
double acti=.0;
double temp=.0;
double ret=.0;
for(int i=0;i<proteinSiteActivations.length;i++){ //over all cis-sites (one in Quick implementation)
  temp=999999.0;                                    
  for(int j=0;j<proteinSiteActivations[i].length;j++){//over all binding sites (AND combo)
    if(proteinSiteActivations[i][j]<temp)             //not relevant for Quick
      temp=proteinSiteActivations[i][j];              //implementation, as we take
    proteinSiteActivations[i][j]*=holder.proteinDecay[geneCisReceptorSites[i][j+1]];
                                    //decay           //smallest value and there is
    }                                                 //only one. Later AND combination.
  //sum up (+ or -) the activations (NOT and OR combos):
  if(geneCisReceptorSites[i][0]==0) //inhibitory
    acti-=temp;
  else                              //ativatory
    acti+=temp;
  }

double shift=0.0;
if(geneOutputActivity==1) //default 'off' gene
  shift=15.0;
else //default 'on' gene
  shift=-5.0;
//def MathLib.tanhScaled(x, range, floor=nil, rshift=0, slope=2)
//System.out.print("acti of "+acti);
if(acti>actiInputCeiling)
  acti=actiInputCeiling;
ret=tanhScaled(acti,stdRange,.0,shift,stdSlope);
//System.out.println(" results in: "+ret);
if(ret<.0) //concentrations can of course never be negative!
  ret=.0;  //(output is between 0 and floor+range, which is currently 0+150=150)
if(Double.isNaN(ret))
  ret=.0;
return ret;
}

/*
returns the number of Cis-sites this gene has for a particular protein
(i.e. the protein number "which" matches and can be docked)
in the Quick implementation this is always 0 or 1
*/
public int getNoOfCisSitesForProt(int which){
int ret=0;
int[] curr;
for(int i=0;i<geneCisReceptorSites.length;i++){
  curr=getCisSite(i);
  for(int j=1;j<curr.length;j++){//Position 0 is activ./inhib. and we don't care here
    if(curr[j]==which)
      ret++;
  }
}
return ret;
}

}
